#include "multiprec.h"
#include <string.h>

/** @addtogroup multiprec
 *  @{
 */
/** @addtogroup arithmetic
 *  @{
 */





/**
 * @addtogroup lowlevel
 * @{
 */


/**
 * digit_tɽƤͤ¤
 *
 * @param result ̤ǼƬؤΥݥ
 * @param lhs 1ɽƬؤΥݥ
 * @param rhs 2ɽƬؤΥݥ
 * @param lhs_len lhsΥݥȤǿ
 * @param rhs_len rhsΥݥȤǿ
 *
 * @pre lhs_len >= rhs_len
 * @pre resultΥݥȤϺlhs_len+1ǤǼǤ뤳ȡ
 * @note lhs, rhs, resultפƤ褤
 *
 * @return result˳Ǽ줿ǿ
 */
inline size_t
ymp_addabs(digit_t *result, const digit_t *lhs, const digit_t *rhs,
	   size_t lhs_len, size_t rhs_len)
{
  double_digit_t tmp = 0;
  size_t i;

  for ( i = 0; i < rhs_len; ++i)
    {
      tmp += lhs[i];
      tmp += rhs[i];
      result[i] = LOW_DIGIT(tmp);
      tmp = HIGH_DIGIT(tmp);
    }
  for ( ; i < lhs_len; ++i)
    {
      tmp += lhs[i];
      result[i] = LOW_DIGIT(tmp);
      tmp = HIGH_DIGIT(tmp);
    }
  if (LOW_DIGIT(tmp))
    {
      result[i++] = LOW_DIGIT(tmp);
    }
  return i;
}



/**
 * digit_tɽƤͤκ
 *
 * @param result ̤ǼƬؤΥݥ
 * @param lhs 1ɽƬؤΥݥ
 * @param rhs 2ɽƬؤΥݥ
 * @param lhs_len lhsΥݥȤǿ
 * @param rhs_len rhsΥݥȤǿ
 *
 * @pre (lhsɽ) >= (rhsɽ)
 * @pre resultΥݥȤϺlhs_lenǤǼǤ뤳ȡ
 * @note lhs, rhs, resultפƤ褤
 *
 * @return result˳Ǽ줿ǿ
 */
inline size_t
ymp_subabs(digit_t *result, const digit_t *lhs, const digit_t *rhs,
	   size_t lhs_len, size_t rhs_len)
{
  double_digit_t tmp = 1;
  size_t i;

  for ( i = 0; i < rhs_len; ++i)
    {
      tmp += MULTIPREC_RADIX - 1;
      tmp += lhs[i];
      tmp -= rhs[i];
      result[i] = LOW_DIGIT(tmp);
      tmp = HIGH_DIGIT(tmp);
    }
  for ( ; i < lhs_len; ++i)
    {
      tmp += MULTIPREC_RADIX - 1;
      tmp += lhs[i];
      result[i] = LOW_DIGIT(tmp);
      tmp = HIGH_DIGIT(tmp);
    }
  while (i-- > 0)
    {
      if (result[i]) break;
    }
  return ++i;
}

/** @} */


/**
 * ø¹Ԥ롣
 *
 * \par is_add0ΤȤ
 *   \li resultselfˤ롣
 *   \li resultͤselfͤotherͤ¤ˤ
 * 
 * \par is_add0ΤȤ
 *   \li resultselfotherΤͤ礭ۤˤ
 *   \li resultͤselfͤotherͤκˤ
 *
 * @pre result, self, otherϽѤΤ
 * @note result, self, otherפƤ褤
 */
inline static void
ymp_do_add(mp_ref_t result, mp_cref_t self, mp_cref_t other, int is_add)
{
  if (self->used > other->used)
    {
      if (is_add)
	{
	  ymp_reserve(result, self->used+1);
	  result->used
	    = ymp_addabs(result->digits, self->digits, other->digits,
			 self->used, other->used);
	  result->sign = self->sign;
	}
      else
	{
	  ymp_reserve(result, self->used);
	  result->used
	    = ymp_subabs(result->digits, self->digits, other->digits,
			 self->used, other->used);
	  result->sign = self->sign;
	}
      return;
    }
  else if (self->used < other->used)
    {
      if (is_add)
	{
	  ymp_reserve(result, other->used+1);
	  result->used
	    = ymp_addabs(result->digits, other->digits, self->digits,
			 other->used, self->used);
	  result->sign = self->sign;
	}
      else
	{
	  ymp_reserve(result, other->used);
	  result->used
	    = ymp_subabs(result->digits, other->digits, self->digits,
			 other->used, self->used);
	  result->sign = !self->sign;
	}
      return;
    }



  /* self->used == other->used */

  if (is_add)
    {
      ymp_reserve(result, self->used+1);
      result->used
	= ymp_addabs(result->digits, self->digits, other->digits,
		     self->used, other->used);
      result->sign = self->sign;
    }
  else 
    {
      size_t i = self->used;
      while (i-- > 0)
	{
	  if ( self->digits[i] > other->digits[i])
	    {
	      ymp_reserve(result, ++i);
	      result->used
		= ymp_subabs(result->digits, self->digits, other->digits,
			     i, i);
	      result->sign = self->sign;
	      return;
	    }
	  else if (self->digits[i] < other->digits[i])
	    {
	      ymp_reserve(result, ++i);
	      result->used
		= ymp_subabs(result->digits, other->digits, self->digits,
			     i, i);
	      result->sign = !self->sign;
	      return;
	    }
	}
      result->sign = positive_sign;
      result->used = 0;
    }
}



/**
 * @ref digit_tȤβø¹Ԥ롣
 *
 * \par is_add0ΤȤ
 *   \li resultselfˤ롣
 *   \li resultͤselfͤotherͤ¤ˤ
 * 
 * \par is_add0ΤȤ
 *   \li resultselfotherΤͤ礭ۤˤ
 *   \li resultͤselfͤotherͤκˤ
 *
 * @pre result, selfϽѤΤ
 * @note self, resultפƤ褤
 */
inline static void
ymp_do_add_digit(mp_ref_t result, mp_cref_t self, digit_t other, int is_add)
{
  switch (self->used)
    {
    case 0:
      ymp_reserve(result, 1);
      result->sign = self->sign;
      result->used = 1;
      result->digits[0] = other;
      return;

    case 1:
      if (is_add)
	{
	  double_digit_t tmp;

	  ymp_reserve(result, 2);
	  result->sign = self->sign;
	  result->used = 1;

	  tmp = other;
	  tmp += self->digits[0];

	  result->digits[0] = LOW_DIGIT(tmp);
	  if ( (tmp = HIGH_DIGIT(tmp)) )
	    {
	      self->digits[1] = LOW_DIGIT(tmp);
	      ++result->used;
	    }
	  return;
	}
      else
	{
	  ymp_reserve(result, 1);
	  result->used = 1;
	  if (self->digits[0] > other)
	    {
	      result->sign = self->sign;
	      result->digits[0] = self->digits[0] - other;
	    }
	  else
	    {
	      result->sign = !self->sign;
	      self->digits[0] = other - self->digits[0];
	    }
	  return;
	}
    default:
      if (is_add)
	{
	  ymp_reserve(result, self->used+1);
	  result->used 
	    = ymp_addabs(result->digits, self->digits, &other, self->used, 1);
	  result->sign = self->sign;
	  return;
	}
      else
	{
	  ymp_reserve(result, self->used+1);
	  result->used
	    = ymp_subabs(result->digits, self->digits, &other, self->used, 1);
	  result->sign = self->sign;
	  return;
	}
    }
}





/**
 * ¿Ĺβû
 *
 * <result><self> + <other>Ǽ
 * @param self ­ɽstruct multiprecؤΥݥ
 * @param other ­ɽstruct multiprecؤΥݥ
 * @param result ̤Ǽstruct multiprecؤΥݥ
 * @pre result, self, otherϽѤΤ
 * @note result, self, otherפƤ褤
 */
void ymp_add(mp_ref_t result, mp_cref_t self, mp_cref_t other)
{
  ymp_do_add(result, self, other, self->sign==other->sign);
}

/**
 * ¿Ĺθ
 *
 * <result><self> - <other>Ǽ
 * @param self ɽstruct multiprecؤΥݥ
 * @param other ɽstruct multiprecؤΥݥ
 * @param result ̤Ǽstruct multiprecؤΥݥ
 * @pre result, self, otherϽѤΤ
 * @note result, self, otherפƤ褤
 */
void ymp_sub(mp_ref_t result, mp_cref_t self, mp_cref_t other)
{
  ymp_do_add(result, self, other, self->sign!=other->sign);
}



/**
 * ¿Ĺ@ref digit_t­
 * @param result ¤Ǽstruct multiprecؤΥݥ
 * @param self ­ɽstruct multiprecؤΥݥ
 * @param val ­
 * @pre result, selfϽѤΤ
 */
void
ymp_add_digit(mp_ref_t result, mp_cref_t self, digit_t val)
{
  ymp_do_add_digit(result, self, val, self->sign==positive_sign);
}

/**
 * ¿Ĺ@ref digit_t
 * @param result Ǽstruct multiprecؤΥݥ
 * @param self ɽstruct multiprecؤΥݥ
 * @param val 
 * @pre result, selfϽѤΤ
 */
void 
ymp_sub_digit(mp_ref_t result, mp_cref_t self, digit_t val)
{
  ymp_do_add_digit(result, self, val, self->sign!=positive_sign);
}


/**
 * ¿Ĺȿž
 *
 * <result> := -<self>
 * @param result ̤Ǽstruct multiprecؤΥݥ
 * @param orig οǼƤstruct multiprecؤΥݥ
 * @pre result, origϽѤΤ
 */
void ymp_neg(mp_ref_t result, mp_cref_t orig)
{
  result->sign = !orig->sign;
  if (result != orig)
    {
      ymp_reserve(result, orig->used);
      result->used = orig->used;
      memcpy(result->digits, orig->digits, sizeof(digit_t)*orig->used);
    }
}

/** @} */
/** @} */

