#include "fft.h"
#include <math.h>

/**
 * @addtogroup fft ®աꥨѴ
 * @{
 */

/**
 * ʣǿξFFTμ¹ʬ
 *
 * @param r_result ̤μ
 * @param i_result ̤ε
 * @param r_source μ
 * @param i_source ε
 * @param len Ĺ(stepñ̤)
 * @param step
 * @pre len2ζ
 */
static void
do_fft(double *r_result, double *i_result, 
       const double *r_source, const double *i_source, 
       size_t len, size_t step)
{
  size_t len_2 = len/2;
  unsigned j;
  if (len == 1) 
    {
      r_result[0] = r_source[0]; i_result[0] = i_source[0];
      return;
    }
  
  do_fft(r_result, i_result, r_source, i_source, 
	 len_2, step*2);
  do_fft(r_result+step, i_result+step, r_source+step, i_source+step, 
	 len_2, step*2);

  for (j=0; j<len_2; ++j)
    {
      double r_xi_j, i_xi_j;
      double r_tmp1 = r_result[j*step*2];
      double i_tmp1 = i_result[j*step*2];
      double r_tmp2 = r_result[j*step*2+step];
      double i_tmp2 = i_result[j*step*2+step];
      
      r_xi_j = cos(M_PI*j/len_2); /* Re(xi_n^j) */
      i_xi_j = sin(M_PI*j/len_2); /* |Im(xi_n^j)| */

      r_result[j*step] = r_tmp1 + r_xi_j*r_tmp2 - i_xi_j*i_tmp2;
      i_result[j*step] = i_tmp1 + r_xi_j*i_tmp2 + i_xi_j*r_tmp2;

      r_xi_j = cos(M_PI*2*(j+len_2)/len); /* Re(xi_n^{j+n/2}) */
      i_xi_j = sin(M_PI*2*(j+len_2)/len); /* |Im(xi_n^{j+n/2})| */

      r_result[(j+len_2)*step] = r_tmp1 + r_xi_j*r_tmp2 - i_xi_j*i_tmp2;
      i_result[(j+len_2)*step] = i_tmp1 + r_xi_j*i_tmp2 + i_xi_j*r_tmp2;
    }
}


/**
 * ʣǿξι®աꥨѴ
 *
 * @param r_result ̤μ
 * @param i_result ̤ε
 * @param r_source μ
 * @param i_source ε
 * @param len Ĺ
 * @pre len2ζ
 */
void
fft(double *r_result, double *i_result, 
    const double *r_source, const double *i_source, 
    size_t len)
{
  do_fft(r_result, i_result, r_source, i_source, len, 1);
} 


/**
 * ʣǿξʣǿξι®եաꥨѴ
 *
 * @param r_result ̤μ
 * @param i_result ̤ε
 * @param r_source μ
 * @param i_source ε
 * @param len Ĺ
 * @pre len2ζ
 */
void
ifft(double *r_result, double *i_result, 
     const double *r_source, const double *i_source, 
     size_t len)
{
  size_t i;
  double *i_source2 = alloca(sizeof(double)*len);
  for (i=0; i<len; ++i) i_source2[i] = -i_source[i];
  do_fft(r_result, i_result, r_source, i_source2, len, 1);  
  for (i=0; i<len; ++i) r_result[i] /= len;
  for (i=0; i<len; ++i) i_result[i] = -i_result[i]/len;
}

/**
 * @}
 */
