#include "multiprec.h"
#include "fft.h"

/** @addtogroup multiprec
 *  @{
 */
/** @addtogroup arithmetic
 *  @{
 */
/**
 * @addtogroup lowlevel
 * @{
 */

/** FFTη̤Ǽ빽¤ */
struct fft_result
{
  /** Ĺ */
  size_t len;
  /** γƼ */
  double *real;
  /** γƵ */
  double *imag;
};
/** struct fft_resultΰʽ */
#define FFT_RESULT_INITIALIZER {0, NULL, NULL}

/**
 * @ref digit_tFFT˳ݤ
 *
 * @param result ̤Ǽ빽¤ΤؤΥݥ
 * @param digits Ѵ٤digit_t
 * @param len Ĺ(ǿ). 2ζҤǤʤɤ
 * @param pow lenĶŬ2Τ٤
 * @post result->real, result->imagˤϡ줾ymp_malloc()ǳݤ줿
 * ΰݥȤ
 */
inline static void
do_digit_fft(struct fft_result *result, const digit_t *digits, 
	  size_t len, size_t pow)
{
  double *r_source, *i_source;
  size_t i;

  result->len = pow;
  YMP_ALLOCATE(result->real, double, pow);
  YMP_ALLOCATE(result->imag, double, pow);

  r_source = alloca(sizeof(double)*pow);
  i_source = alloca(sizeof(double)*pow);

  for (i=0; i<len; ++i) { r_source[i] = digits[i]; i_source[i] = 0; }
  for (   ; i<pow; ++i) { r_source[i] = 0;         i_source[i] = 0; }

  fft(result->real, result->imag, r_source, i_source, len);
}

/**
 * @ref digit_tFFT˳ݤ
 *
 * @param result ̤Ǽ빽¤ΤؤΥݥ
 * @param digits Ѵ٤digit_t
 * @param len Ĺ(ǿ). 2ζҤǤʤɤ
 * @param pow lenĶŬ2Τ٤
 * @post result->real, result->imagˤϡ줾ymp_malloc()ǳݤ줿
 * ΰݥȤ
 */
void
digit_fft(struct fft_result *result, const digit_t *digits, size_t len)
{
  size_t pow = 1;
  while (len > pow) pow<<=1;
  do_digit_fft(result, digits, len, pow);
}

/**
 * IFFT˳ݤ@ref digit_tˤɤ
 *
 * @param digits Ѵη̤Ǽdigit_tƬؤΥݥ
 * @param source FFT̤Ǽ¤ΤؤΥݥ
 * @param len ٤digit_tĹ(ǿ). 2ζҤǤʤɤ
 * @pre lenϡsource->lenʲǤ뤳
 */
void
digit_ifft(digit_t *result, const struct fft_result *source, size_t len)
{
  static const double multiprec_radix = (double)MULTIPREC_RADIX;
  size_t i;
  double *r_result_buf = alloca(sizeof(digit_t)*source->len);
  double *i_result_buf = alloca(sizeof(digit_t)*source->len);  

  ifft(r_result_buf, i_result_buf, source->real, source->imag, source->len);
  for (i=0; i<len; ++i)
    {
      if (r_result_buf[i] >= multiprec_radix)
	{
	  double q = r_result_buf[i]/multiprec_radix;
	  result[i] = (digit_t)(r_result_buf[i]-q*multiprec_radix);
	  r_result_buf[i+1] += q;
	} 
      else
	{
	  result[i] = (digit_t)r_result_buf[i];
	}
    }
}

/** @} */

void ymp_fft_mulabs(digit_t *result, const digit_t *lhs, const digit_t *rhs,
		    size_t lhs_len, size_t rhs_len)
{
  size_t i;
  struct fft_result fft_lhs, fft_rhs;
  size_t len_max = MAX(lhs_len, rhs_len);
  size_t pow = 1;
  while (len_max > pow) pow<<=1;
  pow<<=1;
  
  do_digit_fft(&fft_lhs, lhs, lhs_len, pow);
  do_digit_fft(&fft_rhs, rhs, rhs_len, pow);


  for (i=0; i<lhs_len+rhs_len; ++i)
    {
      double lhs_real = fft_lhs.real[i], lhs_imag = fft_lhs.imag[i];
      fft_lhs.real[i] = lhs_real*fft_rhs.real[i] - lhs_imag*fft_rhs.imag[i];
      fft_lhs.imag[i] = lhs_imag*fft_rhs.real[i] + lhs_real*fft_rhs.imag[i];
    }

  digit_ifft(result, &fft_lhs, lhs_len+rhs_len);
}

/** @} */
/** @} */
