#include <string.h>
#include <assert.h>
#include "multiprec.h"

/** @addtogroup multiprec
 * @{
 */
/** @addtogrup Arithmetic
 * @{
 */
/** @addtogroup finite_fft ͭξι®աꥨѴ
 * @{
 */

void ufft4(digit_t (*result)[3], const digit_t (*source)[3], size_t n);
void uifft4(digit_t (*result)[3], const digit_t (*source)[3], size_t n);

/**
 *
 * @param result ̤ǼƬؤΥݥ
 * @param source FFT-iFFTˤѴ
 * @param len sourceǿ
 * @pre resultϽʬʿǤǼǤ뤳
 * @return result˳Ǽ줿ǿ
 */
size_t do_normalize(digit_t *result, const digit_t (*source)[3], size_t len)
{
  double_digit_t tmp;
  size_t i;
  for (i=0; i<len; ++i)
    {
      
    }
}



/**
 * p = U(5)ˤ6Ĺξ;
 *
 * p = 2^32 * (2^64 - 1) + 1ˤ;
 * @param result ̤ǼƬؤΥݥ
 * @param source 6ĹɽƬؤΥݥ
 * @pre resultΥݥȤϾʤȤ3ǤǼǤ뤳ȡ
 * @note result, sourceϰפƤɤ
 */
// p : 0x ffffffff ffffffff 00000001
inline static void
ufft4_mod_p(digit_t *result, digit_t *source)
{
  double_digit_t tmp;

  /*   r[5]:  r[4]:  4[3]:      r[2]:           r[1]:  r[0]
   * \cong           r[5]: r[4]+r[2]:      r[3]+r[1]:  r[0]
   * \cong                 r[4]+r[2]: r[3]+r[1]+r[5]: r[0]-r[5]   (mod p)
   */
  
  result[0] = source[0];
  
  tmp = source[1]; tmp += source[3];
  result[1] = LOW_DIGIT(tmp);
  tmp = HIGH_DIGIT(tmp);
  
  tmp += source[2]; tmp += source[4];
  result[2] = LOW_DIGIT(tmp);
  tmp = HIGH_DIGIT(tmp);
  
  tmp += source[5];
  if (tmp)
    {
      digit_t r5 = LOW_DIGIT(tmp);
      tmp = 0x100000000 + result[0];
      tmp -= r5;
      result[0] = LOW_DIGIT(tmp);

      if (!HIGH_DIGIT(tmp)) --r5;
      tmp = result[1]; tmp += r5;
      result[1] = LOW_DIGIT(tmp);

      if (HIGH_DIGIT(tmp)) ++result[2];
    }
  
  if (result[2] == 0xffffffff && result[1] == 0xffffffff && result[0] >= 0x1)
    {
      result[2] = 0; result[1] = 0; --result[0];
    }
}

/**
 * Z/pZˤ3Ĺξ軻(p = U(5))
 * @param result ̤Ǽ롣3ĹɽƬؤΥݥ󥿡
 * @param lhs 3ĹɽƬؤΥݥ󥿡
 * @param rhs 3ĹɽƬؤΥݥ󥿡
 * @param tmp ꥢ6ĹɽƬؤΥݥ󥿡
 * @pre result, lhs, rhsϰפƤɤ
 * @pre tmpϤȤפʤ
 */
inline static void
ufft4_mul(digit_t *result, const digit_t *lhs, const digit_t *rhs,
	  digit_t *tmp)
{
  switch (ymp_mulabs(tmp, lhs, rhs, 3, 3))
    {
    case 0: tmp[0] = 0;
    case 1: tmp[1] = 0;
    case 2: tmp[2] = 0;
    case 3: tmp[3] = 0;
    case 4: tmp[4] = 0;
    case 5: tmp[5] = 0;
    case 6: /* nothing */;
    }
  ufft4_mod_p(result, tmp);
}


/**
 * Z/pZθ(2^n)躬(p = U(5))
 *
 * p = 2^32 * (2^64 - 1) + 1ȤZ/pZθ(2^n)躬롣
 * @param xi ̤Ǽ3ĹؤΥݥ
 * @param n ѻؿ
 * @pre 0<= n < 32
 */
inline static void
ufft4_primitive_root(digit_t *xi, size_t n)
{
  digit_t tmp[6];
  /* 11Z/pZθp躬ʤΤǡ
   * 11^t,  where t=2^(32-n) * (2^64 - 1) in Z/pZ 븶Ϻ */
  /* 11^(2^64-1) == 0xb6e2a58fdbf5f190fd9ed2fd (mod p) */

  xi[0] = 0xfd9ed2fd; xi[1] = 0xdbf5f190; xi[2] = 0xb6e2a58f;
  
  n = 32-n;
  while (n-- != 0)
    {
      /* let xi = xi*xi mod p */
      ufft4_mul(xi, xi, xi, tmp);
    }
}


/**
 * ufft4_primitive_root()εո
 *
 * p = 2^32 * (2^64 - 1) + 1ȤZ/pZθ(2^n)躬ǡufft4_primitive_root()֤ΤεոȤʤäƤΤ׻
 * @param xi ̤Ǽ3ĹؤΥݥ
 * @param n ѻؿ
 * @pre 0<= n < 32
 */
inline static void
uifft4_primitive_root(digit_t *xi, size_t n)
{
  digit_t tmp[6];
  /* 11Z/pZθϺ
   * 11^{-t},  where t=2^(32-n) * (2^64 - 1) in Z/pZ 븶Ϻ */
  /* 11^(-(2^64-1)) == 0xd1745d1745d1745c45d1745e (mod p) */

  xi[0] = 0x45d1745e; xi[1] = 0x45d1745c; xi[2] = 0xd1745d17;
  
  n = 32-n;
  while (n-- != 0)
    {
      /* let xi = xi*xi mod p */
      ufft4_mul(xi, xi, xi, tmp);
    }
}


/**
 * Z/pZFFT¹ (p = 2^32 * (2^64 - 1) + 1)
 *
 * @param result Ѵ̤Ǽ롣
 *   3ĹƬؤΥݥ󥿤ǤꡢstepФǥ
 * @param source ѴǼƤ롣
 *   3ĹƬؤΥݥ󥿤ǤꡢstepФǥ
 * @param n result, sourceĹ(ǿ*stepñ̤Ȥ)ѻؿ
 * @param step ˱Ф٤(ǿ)
 * @param xi Z/pZθ2^n躬ɽƬؤΥݥ
 *
 * @pre resultsourceפƤ褤
 * @pre sourceΥݥȤ(stepФ)2^nĤǤǼƤ뤳
 * @pre resultΥݥȤ(stepФ)ʤȤ2^nĤǤǼǤ뤳
 * @pre xi3ĹݥȤƤ뤳
 * @pre step * 2^n == (ѴĹ)
 * @pre 0<= n < 32
 */
static void
do_ufft4(digit_t (*result)[3], const digit_t (*source)[3],
	 size_t n, digit_t *xi, size_t step)
{
  size_t len, i;
  digit_t xi_2[3]; /* xi^2 */
  digit_t tmp[6];

  if (n == 0)
    {
      assert(xi[0]==1 && xi[1]==0 && xi[2]==0);
      memcpy(&result[0], &source[0], sizeof(result[0]));
      return;
    }

  ufft4_mul(xi_2, xi, xi, tmp); /* let xi_2 = xi^2 */
  do_ufft4(result, source, n-1, xi_2, step*2); 
  do_ufft4(result+step, source+step, n-1, xi_2, step*2);

  len = 1<<(n-1);  /* len = (2^n)/2 */
  for (i=0; i<len; ++i)
    {
      digit_t u_low_xi[3], u_high_xi[3];
      ufft4_mul(u_low_xi, result[i*step], xi, tmp);
      ufft4_mul(u_high_xi, result[(i+len)*step], xi, tmp);
      ufft4_mul(result[i*step], result[i*step], u_low_xi, tmp);
      ufft4_mul(result[(i+len)*step], result[(i+len)*step], u_low_xi, tmp);
    }
}



/**
 * Z/pZFFT
 *
 * @param result Ѵ̤Ǽ롣3ĹƬؤΥݥ
 * @param source ѴǼƤ롣3ĹƬؤΥݥ
 * @param n result, sourceĹѻؿ(log_2)
 * @pre resultsourceפƤ褤
 * @pre sourceΥݥȤ2^nĤǤǼƤ뤳
 * @pre resultΥݥȤϾʤȤ2^nĤǤǼǤ뤳
 * @pre 0<= n < 32
 */

void
ufft4(digit_t (*result)[3], const digit_t (*source)[3], size_t n)
{
  digit_t xi[3];
  ufft4_primitive_root(xi, n);
  do_ufft4(result, source, n, xi, 1);
}



/**
 * Z/pZεFFT
 *
 * @param result Ѵ̤Ǽ롣3ĹƬؤΥݥ
 * @param source ѴǼƤ롣3ĹƬؤΥݥ
 * @param n result, sourceĹѻؿ(log_2)
 * @pre resultsourceפƤ褤
 * @pre sourceΥݥȤ2^nĤǤǼƤ뤳
 * @pre resultΥݥȤϾʤȤ2^nĤǤǼǤ뤳
 * @pre 0<= n < 32
 */
void
uifft4(digit_t (*result)[3], const digit_t (*source)[3], size_t n)
{
  size_t i, len;
  digit_t xi[3];      /* ոϺ */
  digit_t pow_inv[6]; /* 2**nεտ */
  digit_t tmp[6];

  uifft4_primitive_root(xi, n);
  do_ufft4(result, source, n, xi, 1);


  /* divide by (2**n)^{-1} */
  /* 0xfffffffeffffffff00000002 * 2**32 == 1 mod p */
  tmp[0] = 0x00000002 << (n-32);
  tmp[1] = LOW_DIGIT(0xffffffff00000002 << (n-32));
  tmp[2] = LOW_DIGIT(0xfffffffeffffffff << (n-32));
  tmp[3] = LOW_DIGIT(0x00000000fffffffe << (n-32));
  tmp[4] = 0;
  tmp[5] = 0;
  ufft4_mod_p(pow_inv, tmp);

  len = 1<<n;
  for (i=0; i<len; ++i)
    {
      ufft4_mul(result[i], result[i], pow_inv, tmp);
    }
}

/**
 * FFTˤ¿Ĺξ軻(4)
 *
 * p = 2^32 * (2^64 - 1) + 1ȤZ/pZFFTѤ롣
 * http://homepage1.nifty.com/~umetani/ims/2001/20010920002/0.htmlˤ롣
 * @param result ̤ǼƬؤΥݥ
 * @param lhs 1ɽƬؤΥݥ
 * @param rhs 2ɽƬؤΥݥ
 * @param len lhs, rhsΥݥȤǿ
 * @pre resultȡlhsrhsϰפʤ
 * @note lhs, rhsϰפƤɤ
 * @return result˳Ǽ줿ǿ
 */
size_t ufft4_mulabs(digit_t *result, const digit_t *lhs, const digit_t *rhs,
		    size_t lhs_len, size_t rhs_len)
{
  digit_t tmp[6];
  digit_t (*new_lhs)[3], (*new_rhs)[3];
  size_t i, len=1, max_len = MAX(lhs_len, rhs_len);
  size_t log2 = 0;

  assert(DIGIT_BIT == 32);
  while (max_len > len) { len<<=1; ++log2; }


  YMP_TEMP_ALLOCATE(new_lhs, digit_t [3], len);
  YMP_TEMP_ALLOCATE(new_rhs, digit_t [3], len);

  for (i=0; i<lhs_len; ++i)
    {
      new_lhs[i][0] = lhs[i];
      new_lhs[i][1] = 0;
      new_lhs[i][2] = 0;
    }
  memset(new_lhs+i, 0, sizeof(new_lhs[i])*(len-i));
  for (i=0; i<rhs_len; ++i)
    {
      new_rhs[i][0] = rhs[i];
      new_rhs[i][1] = 0;
      new_rhs[i][2] = 0;
    }
  memset(new_rhs+i, 0, sizeof(new_rhs[i])*(len-i));  

  
  ufft4(new_lhs, (const digit_t(*)[3])new_lhs, log2);
  ufft4(new_rhs, (const digit_t(*)[3])new_rhs, log2);

  for (i=0; i<len; ++i)
    {
      ufft4_mul(new_lhs[i], new_lhs[i], new_rhs[i], tmp);
    }

  uifft4(new_lhs, (const digit_t(*)[3])new_lhs, log2);

  return do_normalize(result, (const digit_t(*)[3])new_lhs[i], len);
}



/** @} */
/** @} */
/** @} */
